#include "OgreVisibilityTerrainSceneManager.h"
#include "OgreOctreeHierarchyInterface.h"
#include "OgreVisibilityOptionsManager.h"
#include <OgreMath.h>
#include <OgreIteratorWrappers.h>
#include <OgreRenderSystem.h>
#include <OgreCamera.h>
#include <OgreLogManager.h>
#include <OgreStringConverter.h>


namespace Ogre {

//-----------------------------------------------------------------------
VisibilityTerrainSceneManager::VisibilityTerrainSceneManager(
	GtpVisibility::VisibilityManager *visManager): 
mVisibilityManager(visManager), 
mUseDepthPass(false),
mIsDepthPass(false),
mShowVisualization(false),
mRenderNodesForViz(false),
mRenderNodesContentForViz(false),
mVisualizeCulledNodes(false),
mSkipTransparents(false),
mDelayRenderTransparents(true)
{
	mHierarchyInterface = 
		new OctreeHierarchyInterface(this, mDestRenderSystem);

    //mDisplayNodes = true;
	//mShowBoundingBoxes = true;

	// TODO: set maxdepth to reasonable value
	mMaxDepth = 50;
	// create material for depth pass
	CreateDepthPass();
}
//-----------------------------------------------------------------------
void VisibilityTerrainSceneManager::CreateDepthPass()
{
	MaterialPtr depthMat = MaterialManager::getSingleton().getByName("Visibility/DepthPass");

	if (depthMat.isNull())
    {
		// Init
		depthMat = MaterialManager::getSingleton().create(
                "Visibility/DepthPass",
                ResourceGroupManager::DEFAULT_RESOURCE_GROUP_NAME);
        mDepthPass = depthMat->getTechnique(0)->getPass(0);
		mDepthPass->setColourWriteEnabled(false);
		mDepthPass->setDepthWriteEnabled(true);
		mDepthPass->setLightingEnabled(false);
		depthMat->load();
	}
	else
	{
		mDepthPass = depthMat->getTechnique(0)->getPass(0);
	}
}
//-----------------------------------------------------------------------
VisibilityTerrainSceneManager::~VisibilityTerrainSceneManager()
{
	delete mHierarchyInterface;
}
//-----------------------------------------------------------------------
Pass *VisibilityTerrainSceneManager::setPass(Pass* pass)
{
	// setting vertex program is not efficient
	//Pass *usedPass = ((mIsDepthPass && !pass->hasVertexProgram()) ? mDepthPass : pass);		
	Pass *usedPass = (mIsDepthPass ? mDepthPass : pass);	

	if (mIsDepthPass)
	{
		// set vertex program of current pass
        if (pass->hasVertexProgram())
		{
			mDepthPass->setVertexProgram(pass->getVertexProgramName());

			if (mDepthPass->hasVertexProgram())
			{
				const GpuProgramPtr& prg = mDepthPass->getVertexProgram();
				// Load this program if not done already
				if (!prg->isLoaded())
					prg->load();
				// Copy params
				mDepthPass->setVertexProgramParameters(pass->getVertexProgramParameters());
			}
		}
		else if (mDepthPass->hasVertexProgram())
		{
			mDepthPass->setVertexProgram("");
		}
	}

	TerrainSceneManager::setPass(usedPass);

	return usedPass;
}
//-----------------------------------------------------------------------
void VisibilityTerrainSceneManager::ShowVisualization(Camera *cam)
{
	// add player camera for visualization purpose
	try {
		Camera *c;
		if ((c = getCamera("PlayerCam")) != NULL)
		{
			getRenderQueue()->addRenderable(c);
		}   
    }
    catch(...)
    {
        // ignore
    }
	
	if (mRenderNodesForViz || mRenderNodesContentForViz)
	{
		for (NodeList::iterator it = mVisible.begin(); it != mVisible.end(); ++it)
		{
			if (mRenderNodesForViz)
			{
				//getRenderQueue()->addRenderable(*it);
				// addbounding boxes instead of node itself
				(*it)->_addBoundingBoxToQueue(getRenderQueue());
			}
			if (mRenderNodesContentForViz) 
			{
				(*it)->_addToRenderQueue(cam, getRenderQueue(), false);
			}
		}
	}
	for (BoxList::iterator it = mBoxes.begin(); it != mBoxes.end(); ++it)
	{
		getRenderQueue()->addRenderable(*it);
	}
}
//-----------------------------------------------------------------------
void VisibilityTerrainSceneManager::_findVisibleObjects(Camera* cam, bool onlyShadowCasters)
{
	//-- show visible scene nodes and octree bounding boxes from last frame
	if (mShowVisualization)
    {
		ShowVisualization(cam);
	}
	else 
	{
		mVisible.clear();
	    mBoxes.clear();
		
		// if there is no depth pass =>
		// we interleave identification and rendering of objects 
		// in _renderVisibibleObjects

		// only shadow casters will be rendered in shadow texture pass
		mHierarchyInterface->SetOnlyShadowCasters(onlyShadowCasters);
	}
}
//-----------------------------------------------------------------------
void VisibilityTerrainSceneManager::_renderVisibleObjects()
{
	// visualization: apply standard rendering
	if (mShowVisualization)
	{	
		TerrainSceneManager::_renderVisibleObjects();
		return;
	}


	//-- hierarchical culling
	// the objects of different layers (e.g., background, scene, 
	// overlay) must be identified and rendered one after another

	bool leaveTransparentsInQueue = mDelayRenderTransparents && !mUseDepthPass;

	// possible two cameras (one for culling, one for rendering)
	mHierarchyInterface->InitFrame(mOctree, mCameraInProgress, 
							mCullCamera ? getCamera("CullCamera") : NULL,
							leaveTransparentsInQueue);

	// call initframe to reset culling manager stats
	mVisibilityManager->GetCullingManager()->InitFrame(mVisualizeCulledNodes);
	
	mSkipTransparents = false;

	//-- render background, in case there is one
	clearSpecialCaseRenderQueues();
	addSpecialCaseRenderQueue(RENDER_QUEUE_BACKGROUND);
	addSpecialCaseRenderQueue(RENDER_QUEUE_SKIES_EARLY);
	setSpecialCaseRenderQueueMode(SceneManager::SCRQM_INCLUDE);

	SceneManager::_renderVisibleObjects();


#ifdef GTP_VISIBILITY_MODIFIED_OGRE
	_deleteRenderedQueueGroups(false);
#endif

	//-- render visible objects (i.e., all but overlay and skies late)
	clearSpecialCaseRenderQueues();
	addSpecialCaseRenderQueue(RENDER_QUEUE_SKIES_LATE);
	addSpecialCaseRenderQueue(RENDER_QUEUE_OVERLAY);
	setSpecialCaseRenderQueueMode(SceneManager::SCRQM_EXCLUDE);

	// transparents are skipped from hierarchical rendering
	// => they need sorting, thus we render them afterwards
	mSkipTransparents = mDelayRenderTransparents;

	// set state for depth pass
    mIsDepthPass = mUseDepthPass;
	
	/** 
	  * the hierarchical culling algorithm
	  * for depth pass: will just find objects and update depth buffer
	  * for delayed rendering: will render all but transparents
	**/
	
	mVisibilityManager->ApplyVisibilityCulling();


	// for depth pass: add visible nodes found with the visibility culling
	if (mUseDepthPass)
	{
		for (NodeList::iterator it = mVisible.begin(); it != mVisible.end(); ++it)
		{
			(*it)->_addToRenderQueue(mCameraInProgress, getRenderQueue(), false);
		}
		mIsDepthPass = false;
	}
	mSkipTransparents = false;

	//-- now we can render all remaining queue objects
	// for depth pass: all
	// for delayed rendering: transparents, overlay 
	clearSpecialCaseRenderQueues();
	SceneManager::_renderVisibleObjects();

	//WriteLog(); // write out stats
}
//-----------------------------------------------------------------------
void VisibilityTerrainSceneManager::_updateSceneGraph(Camera* cam)
{
	mVisibilityManager->GetCullingManager()->SetHierarchyInterface(mHierarchyInterface);
	mHierarchyInterface->SetRenderSystem(mDestRenderSystem);

#ifdef GTP_VISIBILITY_MODIFIED_OGRE
    mHierarchyInterface->SetNumOctreeNodes(mNumOctreeNodes);
#endif
	TerrainSceneManager::_updateSceneGraph(cam);
}
//-----------------------------------------------------------------------
bool VisibilityTerrainSceneManager::setOption(const String & key, const void * val)
{
	if (key == "UseDepthPass")
	{
		mUseDepthPass = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "ShowVisualization")
	{
		mShowVisualization = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "RenderNodesForViz")
	{
		mRenderNodesForViz = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "RenderNodesContentForViz")
	{
		mRenderNodesContentForViz = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "SkyBoxEnabled")
	{
		mSkyBoxEnabled = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "SkyPlaneEnabled")
	{
		mSkyPlaneEnabled = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "SkyDomeEnabled")
	{
		mSkyDomeEnabled = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "VisualizeCulledNodes")
	{
		mVisualizeCulledNodes = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "DelayRenderTransparents")
	{
		mDelayRenderTransparents = (*static_cast<const bool *>(val));
		return true;
	}
	return VisibilityOptionsManager(mVisibilityManager, mHierarchyInterface).
		setOption(key, val) || TerrainSceneManager::setOption(key, val);
}
//-----------------------------------------------------------------------
bool VisibilityTerrainSceneManager::getOption(const String & key, void *val)
{
	if (key == "NumHierarchyNodes")
	{
		* static_cast<unsigned int *>(val) = (unsigned int)mNumOctreeNodes;
		return true;
	}
	
	return VisibilityOptionsManager(mVisibilityManager, mHierarchyInterface).
		getOption(key, val) && TerrainSceneManager::getOption(key, val);
}
//-----------------------------------------------------------------------
bool VisibilityTerrainSceneManager::getOptionValues(const String & key, StringVector &refValueList)
{
	return TerrainSceneManager::getOptionValues( key, refValueList);
}
//-----------------------------------------------------------------------
bool VisibilityTerrainSceneManager::getOptionKeys(StringVector & refKeys)
{
	return VisibilityOptionsManager(mVisibilityManager, mHierarchyInterface).
		getOptionKeys(refKeys) || TerrainSceneManager::getOptionKeys(refKeys);
}
//-----------------------------------------------------------------------
void VisibilityTerrainSceneManager::setVisibilityManager(GtpVisibility::VisibilityManager *visManager)
{
	mVisibilityManager = visManager;
}
//-----------------------------------------------------------------------
GtpVisibility::VisibilityManager *VisibilityTerrainSceneManager::getVisibilityManager( void )
{
	return mVisibilityManager;
}
//-----------------------------------------------------------------------
void VisibilityTerrainSceneManager::WriteLog()
{
	std::stringstream d;

	d << "Algorithm type: " << mVisibilityManager->GetCullingManagerType() << ", "
	  << "Hierarchy nodes: " << mNumOctreeNodes << ", " 
	  << "Traversed nodes: " << mHierarchyInterface->GetNumTraversedNodes() << ", "
	  << "Rendered nodes: " << mHierarchyInterface->GetNumRenderedNodes() << ", "
	  << "Query culled nodes: " << mVisibilityManager->GetCullingManager()->GetNumQueryCulledNodes() << ", "
	  << "Frustum culled nodes: " << mVisibilityManager->GetCullingManager()->GetNumFrustumCulledNodes() << ", "
      << "Queries issued: " << mVisibilityManager->GetCullingManager()->GetNumQueriesIssued() << "\n";

	LogManager::getSingleton().logMessage(d.str());
}
//-----------------------------------------------------------------------
void VisibilityTerrainSceneManager::renderObjects(const RenderPriorityGroup::TransparentRenderablePassList& objs, 
            bool doLightIteration, const LightList* manualLightList)
{
	if (!mSkipTransparents)
	{
		OctreeSceneManager::renderObjects(objs, doLightIteration, manualLightList);
	}
}

} // namespace Ogre
